@@ -51,6 +51,14 @@ def pytest_addoption(parser):
     group = parser.getgroup(
         "rerunfailures", "re-run failing tests to eliminate flaky failures"
     )
+    group._addoption(
+        "--force-reruns",
+        action="store",
+        dest="force_reruns",
+        type=int,
+        help="Force rerunning all tests the specified number of times,"
+        " irrespective of individual test markers.",
+    )
     group._addoption(
         "--only-rerun",
         action="append",
@@ -85,6 +93,13 @@ def pytest_addoption(parser):
         "regex provided. Pass this flag multiple times to accumulate a list "
         "of regexes to match",
     )
+    group._addoption(
+        "--fail-on-flaky",
+        action="store_true",
+        dest="fail_on_flaky",
+        help="Fail the test run with exit code 7 if a flaky test passes on a rerun.",
+    )
+
     arg_type = "string"
     parser.addini("reruns", RERUNS_DESC, type=arg_type)
     parser.addini("reruns_delay", RERUNS_DELAY_DESC, type=arg_type)
@@ -105,6 +120,10 @@ def _get_marker(item):
 
 
 def get_reruns_count(item):
+    reruns = item.session.config.getvalue("force_reruns")
+    if reruns is not None:
+        return reruns
+
     rerun_marker = _get_marker(item)
     # use the marker as a priority over the global setting.
     if rerun_marker is not None:
@@ -251,26 +270,24 @@ def _get_rerun_filter_regex(item, regex_name):
     return regex
 
 
-def _matches_any_rerun_error(rerun_errors, report):
-    return _try_match_reprcrash(rerun_errors, report)
+def _matches_any_rerun_error(rerun_errors, excinfo):
+    return _try_match_error(rerun_errors, excinfo)
 
 
-def _matches_any_rerun_except_error(rerun_except_errors, report):
-    return _try_match_reprcrash(rerun_except_errors, report)
+def _matches_any_rerun_except_error(rerun_except_errors, excinfo):
+    return _try_match_error(rerun_except_errors, excinfo)
 
 
-def _try_match_reprcrash(rerun_errors, report):
-    for rerun_regex in rerun_errors:
-        try:
-            if re.search(rerun_regex, report.longrepr.reprcrash.message):
-                return True
-        except AttributeError:
-            if re.search(rerun_regex, report.longreprtext):
+def _try_match_error(rerun_errors, excinfo):
+    if excinfo:
+        err = f"{excinfo.type.__name__}: {excinfo.value}"
+        for rerun_regex in rerun_errors:
+            if re.search(rerun_regex, err):
                 return True
     return False
 
 
-def _should_hard_fail_on_error(item, report):
+def _should_hard_fail_on_error(item, report, excinfo):
     if report.outcome != "failed":
         return False
 
@@ -283,24 +300,24 @@ def _should_hard_fail_on_error(item, report):
 
     elif rerun_errors and (not rerun_except_errors):
         # Using --only-rerun but not --rerun-except
-        return not _matches_any_rerun_error(rerun_errors, report)
+        return not _matches_any_rerun_error(rerun_errors, excinfo)
 
     elif (not rerun_errors) and rerun_except_errors:
         # Using --rerun-except but not --only-rerun
-        return _matches_any_rerun_except_error(rerun_except_errors, report)
+        return _matches_any_rerun_except_error(rerun_except_errors, excinfo)
 
     else:
         # Using both --only-rerun and --rerun-except
-        matches_rerun_only = _matches_any_rerun_error(rerun_errors, report)
+        matches_rerun_only = _matches_any_rerun_error(rerun_errors, excinfo)
         matches_rerun_except = _matches_any_rerun_except_error(
-            rerun_except_errors, report
+            rerun_except_errors, excinfo
         )
         return (not matches_rerun_only) or matches_rerun_except
 
 
 def _should_not_rerun(item, report, reruns):
     xfail = hasattr(report, "wasxfail")
-    is_terminal_error = _should_hard_fail_on_error(item, report)
+    is_terminal_error = item._terminal_errors[report.when]
     condition = get_reruns_condition(item)
     return (
         item.execution_count > reruns
@@ -344,8 +361,24 @@ class XDistHooks:
         db = sched.config.failures_db
         reruns = db.get_test_reruns(crashitem)
         if db.get_test_failures(crashitem) < reruns:
-            sched.mark_test_pending(crashitem)
-            report.outcome = "rerun"
+            try:
+                sched.mark_test_pending(crashitem)
+                report.outcome = "rerun"
+            except NotImplementedError:
+                # Some schedulers (like LoadScopeScheduling) don't implement
+                # mark_test_pending
+                # In this case, we can't reschedule the crashed test for rerun
+                # Mark it as failed with a clear message about why it couldn't be rerun
+                report.outcome = "failed"
+                if not hasattr(report, "longrepr") or report.longrepr is None:
+                    error_msg = (
+                        "Test crashed and could not be rescheduled for rerun."
+                        f" The scheduler '{sched.__class__.__name__}' does not support"
+                        " rescheduling crashed tests"
+                        " (mark_test_pending not implemented)."
+                        f" Remaining reruns: {reruns - db.get_test_failures(crashitem)}"
+                    )
+                    report.longrepr = error_msg
 
         db.add_test_failure(crashitem)
 
@@ -416,7 +449,7 @@ class SocketDB(StatusDB):
 class ServerStatusDB(SocketDB):
     def __init__(self):
         super().__init__()
-        self.sock.bind(("localhost", 0))
+        self.sock.bind(("127.0.0.1", 0))
         self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
 
         self.rerunfailures_db = {}
@@ -458,7 +491,7 @@ class ServerStatusDB(SocketDB):
 class ClientStatusDB(SocketDB):
     def __init__(self, sock_port):
         super().__init__()
-        self.sock.connect(("localhost", sock_port))
+        self.sock.connect(("127.0.0.1", sock_port))
 
     def _set(self, i: str, k: str, v: int):
         self._sock_send(self.sock, "|".join(("set", i, k, str(v))))
@@ -523,8 +556,9 @@ def pytest_runtest_makereport(item, call):
     _test_failed_statuses = getattr(item, "_test_failed_statuses", {})
     _test_failed_statuses[result.when] = result.failed
     item._test_failed_statuses = _test_failed_statuses
-
-    item._terminal_errors[result.when] = _should_hard_fail_on_error(item, result)
+    item._terminal_errors[result.when] = _should_hard_fail_on_error(
+        item, result, call.excinfo
+    )
 
 
 def pytest_runtest_protocol(item, nextitem):
@@ -614,3 +648,18 @@ def show_rerun(terminalreporter, lines):
         for rep in rerun:
             pos = rep.nodeid
             lines.append(f"RERUN {pos}")
+
+
+@pytest.hookimpl(trylast=True)
+def pytest_sessionfinish(session, exitstatus):
+    if exitstatus != 0:
+        return
+
+    if session.config.option.fail_on_flaky:
+        for item in session.items:
+            if not hasattr(item, "execution_count"):
+                # no rerun requested
+                continue
+            if item.execution_count > 1:
+                session.exitstatus = 7
+                break
